/* $Id: WidgetTest.cpp,v 1.10 2002/11/16 16:40:02 mrq Exp $
**
** Ark - Libraries, Tools & Programs for MMORPG developpements.
** Copyright (C) 1999-2002 The Contributors of the Ark Project
** Please see the file "AUTHORS" for a list of contributors
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sstream>

#include <Ark/ArkTimer.h>
#include <Ark/ArkRenderer.h>
#include <Ark/ArkSystem.h>
#include <Ark/ArkFileSys.h>

#include <Client/WidgetTest.h>
#include <Client/LabelWidget.h>

namespace Client
{

   WidgetTest::WidgetTest (int argc, char **argv)
   {
      Ark::Config *cfg = Ark::Sys()->Cfg();
      
      m_Width = Ark::Sys()->Cfg()->GetInt ("renderer::Width", 640);
      m_Height = Ark::Sys()->Cfg()->GetInt ("renderer::Height", 480);
      
      //-------------------------------------------------------------
      m_Done = false;
  
      if (SDL_Init (SDL_INIT_VIDEO) < 0)
      {
	 Ark::Sys()->Fatal ("Couldn't initialize SDL: %s\n",
			    SDL_GetError ());
      }

      Ark::Sys()->AtExit ((Ark::ExitFunc) &SDL_Quit);
      const SDL_VideoInfo *vi = SDL_GetVideoInfo ();
      int bpp = Ark::Sys()->Cfg()->GetInt ("renderer::BPP",
					   vi->vfmt->BitsPerPixel);

      //--------------------------------------------------------------
      SDL_GL_SetAttribute (SDL_GL_DEPTH_SIZE, 16);
      SDL_GL_SetAttribute (SDL_GL_DOUBLEBUFFER, 1);

      if (SDL_SetVideoMode (m_Width, m_Height, bpp, SDL_OPENGL) == NULL)
      {
	 Ark::Sys()->Fatal ("Couldn't set GL mode: %s", SDL_GetError());
      }

      // Set title
      Ark::String title = cfg->GetStr ("client::Title", "Ark Client");
      SDL_WM_SetCaption (title.c_str(), "arkclient");

      // Display some informations about OpenGL
      Ark::Sys()->Log ("Renderer informations: \n");
      Ark::Sys()->Log ("  Resolution: %dx%dx%d\n", m_Width, m_Height, bpp);
      Ark::Sys()->Log ("  GL_VENDOR: %s\n", glGetString(GL_VENDOR));
      Ark::Sys()->Log ("  GL_VERSION: %s\n", glGetString(GL_VERSION));
      Ark::Sys()->Log ("  GL_RENDERER: %s\n", glGetString(GL_RENDERER));
      Ark::Sys()->Log ("----------\n\n");

      m_Renderer = Ark::RendererFactory::CreateRenderer
	 ("ark::Renderer::OpenGL", 0);

      m_UI = new UIRenderer (m_Renderer, 30);

   }

   WidgetTest::~WidgetTest ()
   {
      if (m_UI)
	 delete m_UI;

      if (m_Renderer)
	 m_Renderer->Unref();
   }

   bool
   WidgetTest::Refresh ()
   {
      glClearColor (0.0, 0.0, 0.0, 1.0);
      glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

      m_Renderer->SetViewport (0, 0, m_Width, m_Height);
      m_CurWidget->Render (m_UI);
      return true;
   }

   bool
   WidgetTest::HandleEvent (SDL_Event *event)
   {
      switch (event->type)
      {
	 case SDL_ACTIVEEVENT:
	    break;

	 case SDL_KEYDOWN:
	    if (event->key.keysym.sym == SDLK_ESCAPE)
	    {
	       m_Done = true;
	       return true;
	    }
	    else if (event->key.keysym.sym == SDLK_F12)
	    {
	       Ark::Image img ("screen", m_Width, m_Height,
			       Ark::Image::RGB_888);

	       /* read our image data from the frame buffer */
	       glReadPixels (0, 0,
			     m_Width, m_Height,
			     GL_RGB, GL_UNSIGNED_BYTE, img.m_Data);

	       Ark::String name;
	       int i = 0;

	       do
	       {
		   std::ostringstream os;
		   os << "{home}/screen" << i << ".tga";
		   name = os.str();
		   ++i;
	       } while (Ark::Sys()->FS()->IsFile (name));

	       img.SaveTGA (name);
	    }
	    m_CurWidget->HandleKey (true,
				    event->key.keysym.mod,
				    event->key.keysym.sym);
	    break;

	 case SDL_KEYUP:
	    m_CurWidget->HandleKey (false,
				    event->key.keysym.mod,
				    event->key.keysym.sym);
	    break;

	 case SDL_QUIT:
	    m_Done = true;
	    break;

	 case SDL_MOUSEBUTTONDOWN:
	    m_CurWidget->HandleMButton (true, event->button.button,
                                        event->button.x,
                                        event->button.y);
	    break;

	 case SDL_MOUSEBUTTONUP:
	    m_CurWidget->HandleMButton (false, event->button.button,
					event->button.x,
					event->button.y);
	    break;
      }

      return true;
   }

   bool
   WidgetTest::ProcessEvents ()
   {
      SDL_Event event;

      int x,y,state = SDL_GetMouseState (&x, &y);
      HandleMotion (state, x, y);

      // Check if there's a pending event.
      while (SDL_PollEvent (&event))
	 HandleEvent(&event);

      return true;
   }

   int
   WidgetTest::Loop ()
   {
      LabelWidget *lw = new LabelWidget("This is a label test !!");
      lw->SetPosition (10, 10);
      m_CurWidget = lw;
      
      while (!m_Done)
      {
	 Refresh ();     
	 SDL_GL_SwapBuffers();
     
	 //-----------------------------------------------------------
	 // Check for SDL/OpenGL error
	 GLenum gl_error = glGetError ();
	 if (gl_error != GL_NO_ERROR)
	 {
	    Ark::Sys()->Log ("OpenGL error: %s (%x)\n",
			     gluErrorString(gl_error), gl_error);
	 }
     
	 char *sdl_error = SDL_GetError ();
	 if (sdl_error[0] != '\0')
	 {
	    Ark::Sys()->Log ("SDL error: %s\n", sdl_error );
	    SDL_ClearError ();
	 }
     
	 //-----------------------------------------------------------
	 ProcessEvents ();
      }

      return 0;
   }


}

int main (int argc, char **argv)
{
  Ark::System::Init (&argc, &argv);
  Ark::Math::Init ();

  Client::WidgetTest wt(argc, argv);

  wt.Loop ();
  return (0);
}
