/* $Id: HFQuadtree.cpp,v 1.5 2002/12/27 11:56:00 mrq Exp $
**
** Ark - Libraries, Tools & Programs for MMORPG developpements.
** Copyright (C) 1999-2002 The Contributors of the Ark Project
** Please see the file "AUTHORS" for a list of contributors
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <algorithm>
#include <math.h>

#ifdef WIN32
 #include <windows.h>
#endif

#include <GL/gl.h>

#include <Ark/ArkRenderer.h>
#include <Modules/HeightField/HFWorld.h>
#include <Modules/HeightField/HFQuadtree.h>


namespace Ark
{

   QuadtreeNode::QuadtreeNode() 
   {
      for (int i = 0; i < 4; i++)
	 m_Children[i] = 0;
   }

   QuadtreeNode::~QuadtreeNode ()
   {
      for (int i = 0; i < 4; i++)
	 delete m_Children[i];
   }

   void
   QuadtreeNode::Build(HeightField* hf, int minX, int minZ, int size)
   {
       const int childSize = size/2;
       const bool usePatch = (childSize == Patch::SIZE);

       for (int i=0 ; i<4 ; ++i)
       {
	   if (usePatch)
	   {
	       m_Children[i] = new Patch();
	   }
	   else
	   {
	       m_Children[i] = new QuadtreeNode();
	   }

	   const int childEast = (i % 2) ? childSize : 0;
	   const int childSouth = (1 < i) ? childSize : 0;
	   const int childX = minX + childEast;
	   const int childZ = minZ + childSouth;

	   // Virtual call to Build
	   m_Children[i]->Build(hf, childX, childZ, childSize);

	   // Make bounding box
	   m_BBox.AddBBox (m_Children[i]->m_BBox);
       }
       // std::cerr << "Node BB is " << m_BBox.m_Min << " " << m_BBox.m_Max << std::endl;

   }

   Patch*
   QuadtreeNode::FindPatch(scalar x, scalar z)
   {
       const Vector3& bbmin = m_BBox.m_Min;
       const Vector3& bbmax = m_BBox.m_Max;

       if ((x < bbmin.X || x > bbmax.X) ||
	       (z < bbmin.Z || z > bbmax.Z))
	   return 0;

       // Leaf is a patch
       if (!m_Children[0])
       {
	   return static_cast<Patch*>(this);
       }

       const scalar midX = (bbmin.X + bbmax.X) / 2.0f;
       const scalar midZ = (bbmin.Z + bbmax.Z) / 2.0f;

       int child = 0;
       if (midX < x)
	   ++child;

       if (midZ < z)
	   child += 2;
       
       return m_Children[child]->FindPatch(x, z); 
   }

   void
   QuadtreeNode::CollectColliders(ColliderList& list)
   {
      for (int i=0 ; i<4 ; ++i)
      {
	 m_Children[i]->CollectColliders(list);
      }
   }

   void
   QuadtreeNode::InvalidatePatches(const BBox& box)
   {
       if (!m_BBox.Overlap(box))
	   return;

       // Leaf is a patch
       if (!m_Children[0])
       {
	   Patch* patch = static_cast<Patch*>(this);
	   patch->m_Dirty = true;
       }
       else
       {
	   for (int i=0 ; i<4 ; ++i)
	   {
	       m_Children[i]->InvalidatePatches(box);
	   }
       }
   }

   bool QuadtreeNode::RayTrace (
			    const Ray& ray,
                            std::vector< int >& potentialquads,
                            EntityList& potentialents)
   {
      Vector3 c;
      if (ray.HitBBox (m_BBox, &c) == false)
	 return false;
	 
      if (m_Children[0])
      {
	 for (int i = 0; i < 4; i++)
	 {
	    if (m_Children[i]->RayTrace (ray, potentialquads, potentialents))
	       return true;
	 }
	 
	 return false;
      }
      else
      {
	  Patch* patch = static_cast<Patch*>(this);
	  patch->AddPatchEntsToList (potentialents);
	  patch->RayTrace(ray, potentialquads);
      }

      return false;
   }

   void Patch::RayTrace(const Ray& ray, std::vector< int >& potentialquads)
   {
       // Needed for boundary checks in patch space
       const Vector3& bbmin = m_BBox.m_Min;
       const scalar size = static_cast<scalar>(SIZE);

       Vector3 v0 = m_Heightfield->GetVectorInGridSpace( ray.m_From ); 
       Vector3 v1 = m_Heightfield->GetVectorInGridSpace( ray.m_To );

       scalar p0[2] = { v0.X - bbmin.X, v0.Z - bbmin.Z };
       scalar p1[2] = { v1.X - bbmin.X, v1.Z - bbmin.Z };

       // Corner for the quad
       // -1 if not found
       int cornerIndex = -1; 

       // Flipping of axis
       bool transform[2] = { false, false }; 
       int which = 0;

       // Patch index moves
       int index = 0;
       int offset[2] = { 1, SIZE + 1 };
       
       // Horizontal shift
       if (p1[0] < p0[0])
       {
	   transform[0] = true;
	   p0[0] = size - p0[0];
	   p1[0] = size - p1[0];
	   index += SIZE;
	   offset[0] = -offset[0];
       }
       
       // Vertical shift
       if (p1[1] < p0[1])
       {
	   transform[1] = true;
	   p0[1] = size - p0[1];
	   p1[1] = size - p1[1];
	   index += (SIZE + 1) * SIZE;
	   offset[1] = -offset[1];
       }

       // First bissectrice shift
       if ((p1[0] - p0[0]) < (p1[1] - p0[1]))
       {
	   which = 1;
       }

       const int other = 1-which;
       const int right = offset[which];
       const int up = offset[other];

       // Select interval
       const scalar minX = p0[which];
       const scalar maxX = p1[which];

       // Length along the abscisses
       const scalar lengthX = maxX - minX;

       // Height of the ray on this scale
       const scalar startHeight = ray.m_From.Y;
       const scalar endHeight = ray.m_To.Y;
       const scalar diffHeight = endHeight - startHeight;
       const scalar diffHeightPerX = diffHeight / lengthX;

       // Length along the ordonnees
       const scalar minY = p0[other];
       const scalar lengthY = p1[other] - minY;
       const scalar slope = lengthY / lengthX;

       // Start position must be minX rounded down or 0
       scalar startX = std::max(floorf(minX), 0.0f);

       // We stay in 1 grid cell (endX-startX) is always bigger than (endY-startY)
       if ((maxX - startX) < 1.0f)
       {
	   // Push the corner and let the end function resolve the intersection
	   const int theX = static_cast<int>( v0.X );
	   const int theZ = static_cast<int>( v0.Z );
	   // std::cerr << "Vertical ray at " << theX << " " << theZ << std::endl;
	   potentialquads.push_back(theX);
	   potentialquads.push_back(theZ);
	   return;
       }

       // Look at Y position there
       scalar lastY = (startX - minX) * slope + minY;
       scalar floorY = floorf(lastY);

       // Simple counter, must be > 1  to have intersection
       // If we hit under the patch on the first test, we did not intersect
       int testCount = 0;

       // We are not yet in the patch
       if (lastY < 0.0f)
       {
	   // Recompute nearest integer X for Y at 0
	   const scalar shiftedX = (0.0f - minY) / slope + minX;
	   assert((startX < shiftedX) && "The new startX should be greater");
	   startX = floorf(shiftedX);
	   
	   // Compute exact Y for this startX
	   lastY = (startX - minX) * slope + minY;
	   floorY = -1.f;
       }

       // shift by firstIndex difference
       const int startIndex = static_cast<int>(startX);
       index += startIndex * right;
       
       // rayHeuight, get increased by diffHeightPerX at each iteration
       scalar currentRayHeight = startHeight + (startX - minX) * diffHeightPerX;
       
       if (0.0f <= lastY)
       {
	   // We have to manually check the first segment
	   // on  the left side of the ray or patch
	   index += static_cast<int>(floorY) * up;

	   const Vector3& vertex0 = m_VB.Coord(index);
	   const Vector3& vertex1 = m_VB.Coord(index + up);
	   const scalar interp = (lastY - floorY) / 1.0f;
	   const scalar height = interp * vertex1.Y + (1.f - interp) * vertex0.Y;

	   if (currentRayHeight < height)
	   {
	       // We are already under the patch, return doing nothing
	       return;
	   }

	   ++testCount;
	}
       
       const scalar endX = std::min(ceilf(maxX), size);
       const int endIndex = static_cast<int>(endX);

       // The computation is cranked one rank to the right
       for (int i=startIndex ; i<endIndex ; ++i)
       {
	   // Advancement ratio
	   const scalar x = static_cast<scalar>(i);
	   const scalar y = (x - minX) * slope + minY;
	   const scalar currentY = floorf(y);

	   // Goes one right
	   index += right;
	   const Vector3& vertex0 = m_VB.Coord(index);

	   // We went one line up
	   if (floorY < currentY)
	   {
	       // Goes up one level
	       index += up;

	       // Check against cell on the left
	       const Vector3& vertexL = m_VB.Coord(index - right);
	       const scalar interp = (currentY - lastY) / (y - lastY);
	       const scalar height = interp * vertex0.Y + (1.f - interp) * vertexL.Y;

	       const scalar rayHeight = startHeight + interp * diffHeightPerX;
	       if (rayHeight < height)
	       {
		   // We got under the patch, break from loop
		   cornerIndex = index - right;
		   break;
	       }

	       ++testCount;
	       if (size <= currentY)
	       {
		   // We went out of the patch
		   break;
	       }

	       floorY = currentY;
	   }
	   
	   // Check against cell on the top
	   currentRayHeight += diffHeightPerX;

	   const Vector3& vertexU = m_VB.Coord(index + up);
	   const scalar interp = (y - currentY) / 1.0f;
	   const scalar height = interp * vertexU.Y + (1.f - interp) * vertex0.Y;

	   if (currentRayHeight < height)
	   {
	       // We got under the patch, break from loop
	       cornerIndex = index - right;
	       break;
	   }

	   ++testCount;
	   lastY = y;
       }
       
       // We got something
       if ((0 <= cornerIndex) && (1 < testCount))
       {
	   // Now we got the index, we must move it to the NorthWest side of the patch
	   if (transform[1])
	       cornerIndex -= SIZE+1;
	   
	   if (transform[0])
	       cornerIndex -= 1;

	   const int row = cornerIndex % (SIZE+1);
	   const int line = cornerIndex / (SIZE+1);
	   
	   potentialquads.push_back(m_MinX + row);
	   potentialquads.push_back(m_MinZ + line);
	   // std::cerr << "Intersection at " << (m_MinX + row) << " and " << (m_MinZ + line) << std::endl;
       }
      
   }


   // =======================================================================
   // Quadtree patch.
   // =======================================================================

   Patch::Patch() : 
       QuadtreeNode(),
       m_Black("black"),
       m_Dirty (true),
       m_CollisionDirty (true)

   {
   }

   Patch::~Patch ()
   {}

   /* The terrain is meshed as follows:
            ...
         +-+-+-+-+            x<---+
         |\|/|\|/|                 |
      ...+-+-+-+-+...              V
         |/|\|/|\|                 z
         +-+-+-+-+
            ...
      
      So there are two types of vertices: those surrounded by
      four triangles (x+z is odd), and those surrounded by
      eight (x+z is even).
   */
   Vector3
   Patch::ComputeNormal (int x, int z, const Vector3& coord) const
   {
      const int nx = m_Heightfield->m_SizeX;
      const int nz = m_Heightfield->m_SizeZ;

      // Border gets Vertical Normal
      if (x < 1 || z < 1 || (nx-1) < x || (nz-1) < z)
      {
	 return Vector3(0.0f, 1.0f, 0.0f);
      }

      Vector3 p0 = m_Heightfield->GetCoord (x, z);
      Vector3 pW = m_Heightfield->GetCoord (x-1, z);
      Vector3 pN = m_Heightfield->GetCoord (x, z-1);
      Vector3 pE = m_Heightfield->GetCoord (x+1, z);
      Vector3 pS = m_Heightfield->GetCoord (x, z+1);
      
      Vector3 normal;
      normal += Vector3::ComputeNormal (p0, pW, pN);
      normal += Vector3::ComputeNormal (p0, pN, pE);
      normal += Vector3::ComputeNormal (p0, pE, pS);
      normal += Vector3::ComputeNormal (p0, pS, pW);
      normal.Normalize();

      return normal;
   }

   void
   Patch::Validate()
   {
      const int maxX = m_MinX + SIZE;
      const int maxZ = m_MinZ + SIZE;
      const int sx = (int) m_Heightfield->m_SizeX - 1;
      const int sz = (int) m_Heightfield->m_SizeZ - 1;

      int idx = 0;
      for (int z = m_MinZ ; z <= maxZ ; z++)
      {
	  for (int x = m_MinX ; x <= maxX ; x++)
	  {
	      const int rx = (sx < x) ? sx : x;
	      const int rz = (sz < z) ? sz : z;

	      Vector3 point = m_Heightfield->GetCoord(rx, rz);

	      // Adds point to bounding box
	      m_BBox.AddPoint(point);

	      m_VB.Coord(idx) = point;
	      m_VB.Normal(idx) = ComputeNormal (rx, rz, m_VB.Coord(idx));
	      Material *grd = m_Heightfield->GetGrd (rx, rz);
	      RGBA &rgba = m_VB.Color4(idx);

	      if (grd)
	      {
		  rgba.R = uchar (grd->m_Diffuse.R * 255.f);
		  rgba.G = uchar (grd->m_Diffuse.G * 255.f);
		  rgba.B = uchar (grd->m_Diffuse.B * 255.f);
		  rgba.A = 255;
	      }
	      else
	      {
		  rgba.R = rgba.G = rgba.B = rgba.A = 255;
	      }

	      ++idx;
	  }
      }

      m_Dirty = false;
      // std::cerr << "Patch BB is " << m_BBox.m_Min << " " << m_BBox.m_Max << std::endl;
   }

   void
   Patch::Build(HeightField* hf, int minX, int minZ, int size)
   {
      assert(size == SIZE);
      assert(hf);

      m_Heightfield = hf;
      m_MinX = minX;
      m_MinZ = minZ;

      // Sets patch offset
      m_Offset = hf->GetCoord(minX, minZ);

      m_VB.SetFormat (VertexBuffer::VB_HAS_COORD|VertexBuffer::VB_HAS_NORMAL|VertexBuffer::VB_HAS_COLOR);
      m_VB.Resize ((SIZE+1) * (SIZE+1));

      Validate();
   }

   uchar
   Patch::Ground (int x, int z) const
   {
      return m_Heightfield->GetGround(x, z);
   }

   int
   Patch::Index (int x, int z) const
   {
      assert (x-m_MinX <= SIZE);
      assert (0 <= x-m_MinX);
      assert (z-m_MinZ <= SIZE);
      assert (0 <= z-m_MinZ);

      return (x-m_MinX) + (z-m_MinZ) * (SIZE+1);
   }

   // Add an entity to this patch..
   void
   Patch::AddEntity (Entity *ent)
   {
      m_CollisionDirty = true;

      EntityLI iterator = std::find(m_Entities.begin(), m_Entities.end(), ent);

      if (iterator == m_Entities.end())
      {
	  m_Entities.push_back(ent);
      }
      
   }
  

   // Remove an entity from this patch..
   void
   Patch::RemoveEntity (Entity *ent)
   {
      m_CollisionDirty = true;

      // Remove entity from list
      m_Entities.erase(std::remove(m_Entities.begin(), m_Entities.end(), ent),
		       m_Entities.end());
   }

   /*
    * Add all the patches entities to the provided list, checking that no
    * entity is added twice in the list.
    */
   void
   Patch::AddPatchEntsToList(EntityList& list)
   {

       for (EntityLI i = m_Entities.begin(); i != m_Entities.end(); ++i)
       {
	   EntityLI found = std::find(list.begin(), list.end(), *i);

	   // Add *i to list if not already found
	   if (found == list.end())
	   {
	       list.push_back (*i);
	   }
       }      
   }

   /**
    * Make a list of potential colliders. If nothing has moved
    * in the patch, we assume there can't be any collision.
    * Note: we use a map to ensure there no collision is computed
    * twice.
    * Once this function has returned, the patch collision dirty
    * flag is cleared.
    */
   void
   Patch::CollectColliders (ColliderList &list)
   {
      if (m_CollisionDirty == false)
	 return;

      for (EntityLI iter1=m_Entities.begin() ; iter1!=m_Entities.end(); ++iter1)
      {
	 Entity* entity1 = *iter1;
	 
	 if (!(entity1->m_Flags & Entity::COLLISION))
	    continue;

	 for (EntityLI iter2=iter1 + 1 ; iter2!=m_Entities.end(); ++iter2)
	 {
	    Entity *entity2 = *iter2;

	    if (!(entity2->m_Flags & Entity::COLLISION) ||
		((entity1->m_Flags & Entity::STATIC) && 
		 (entity2->m_Flags & Entity::STATIC)))
	       continue;

	    if (entity1->GetBBox().Overlap (entity2->GetBBox()))
	    {
	       list[ Collider(entity1,entity2) ] = true;
	    }
	 }
      }
   }

   //========================================================================

   
   Quadtree::Quadtree (HeightField *heightfield) : 
      m_Heightfield (heightfield),
      m_TriBin (heightfield->m_Grounds),
      m_Rootnode (NULL)
   {
      m_Rootnode = new QuadtreeNode();
      m_Size = (int) heightfield->m_SizeX;
      m_Rootnode->Build(heightfield, 0, 0, m_Size);
   }

   Quadtree::~Quadtree ()
   {
      delete m_Rootnode;
   }

   /**
    * Mark the patches in the specified area as "invalidate", which
    * means their topology has changed since the quadtree has been
    * created.
    */
   void
   Quadtree::Invalidate (scalar minx, scalar minz, scalar maxx, scalar maxz)
   {
       const BBox& quadtreeBox = m_Rootnode->GetBoundingBox();
       BBox invalidateBox;
       invalidateBox.m_Min = Vector3(minx, quadtreeBox.m_Min.Y, maxx);
       invalidateBox.m_Max = Vector3(maxx, quadtreeBox.m_Max.Y, maxx);

       m_Rootnode->InvalidatePatches(invalidateBox);
   }

   bool Quadtree::RayTrace (const Ray &ray,
			    Collision &collision,
			    bool test_ents)
   {
       // std::cerr << "RayTrace called" << std::endl;
     std::vector< int > potentialquads;
     EntityList potentialents;

      m_Rootnode->RayTrace (ray, potentialquads, potentialents);

      HeightField *hf = m_Heightfield;

      ColSystem *cs = hf->GetCache()->GetColSystem();

      if (cs != NULL && test_ents)
      {
	 Vector3 p;

	 for (EntityLI e = potentialents.begin(); e != potentialents.end(); ++e)
	 {
	    if (!ray.HitBBox ((*e)->GetBBox(), &p))
	       continue;

	    if (cs->RayTrace ((*e)->m_MState, ray, collision))
	    {
	       collision.m_Flags |= Collision::ENTITY;
	       collision.m_Entity = *e;
	       return true;
	    }
	 }
      }

      Vector3 coord;
      // One quad is a couples (x,z) in hf indices
      const int quadCount = potentialquads.size() / 2;
      for (int i=0; i<quadCount; ++i)
      {
	  const int base = 2*i;
	  const int x = potentialquads[base];
	  const int z = potentialquads[base+1];
	  
	  // Gets the corners
	  Vector3 v[4] = {
	      hf->GetCoord(x, z),
	      hf->GetCoord(x+1, z),
	      hf->GetCoord(x, z+1),
	      hf->GetCoord(x+1, z+1),
	  };

	  int indices[4] = { 0, 1, 2, 3 };
	  
	  // if x+z is odd, the triangles do not break on the base of the quad
	  if ((x+z) % 2)
	  {
	      indices[0] = 1;
	      indices[1] = 0;
	      indices[2] = 3;
	      indices[3] = 2;
	  }

	  if (ray.HitTriangle(v[indices[0]], v[indices[1]], v[indices[2]], &coord))
	  {
	      collision.m_Flags = Collision::WORLD
		  |Collision::PLANE
		  |Collision::POSITION
		  |Collision::WORLD;

	      collision.m_Pos = coord;
	      collision.m_Plane = Plane::GetTriPlane (v[indices[0]], v[indices[1]], v[indices[2]]);
	      collision.m_Material = hf->GetGrd (x, z);
	      // std::cerr << "Intersection at << " << (x) << " and " << (z) << " confirmed" << std::endl;
	      return true;
	  }
	  else if (ray.HitTriangle(v[indices[1]], v[indices[2]], v[indices[3]], &coord))
	  {
	      collision.m_Flags = Collision::WORLD
		  |Collision::PLANE
		  |Collision::POSITION
		  |Collision::WORLD;

	      collision.m_Pos = coord;
	      collision.m_Plane = Plane::GetTriPlane (v[indices[1]], v[indices[2]], v[indices[3]]);
	      collision.m_Material = hf->GetGrd (x, z);
	      // std::cerr << "Intersection at << " << (x) << " and " << (z) << " confirmed" << std::endl;
	      return true;
	  }
      }

      return false;
   }

   // Remove an entity from the quadtree's patches...
   void
   Quadtree::RemoveEntity (Entity *ent, HFEntityData *data)
   {
     std::vector< Patch* >::iterator i;

      for (i = data->m_Patches.begin(); i != data->m_Patches.end(); i++)
	 (*i)->RemoveEntity (ent);
   }
   
   // Add/update an entity in the quadtree's patches.
   void
   Quadtree::UpdateEntity (Entity *ent, HFEntityData *data)
   {
      RemoveEntity (ent, data);

      const BBox &bbox = ent->GetBBox();

      /// Add the entity to all the patches it might belong to...
      scalar cs[4][2] = 
	 {
	    {
	       bbox.m_Min.X,
	       bbox.m_Min.Z,
	    },
	    {
	       bbox.m_Min.X,
	       bbox.m_Max.Z,
	    },
	    {
	       bbox.m_Max.X,
	       bbox.m_Max.Z,
	    },
	    {
	       bbox.m_Max.X,
	       bbox.m_Min.Z,
	    },
	 };

      for (int i = 0; i < 4; i++)
      {
	 Patch *p = m_Rootnode->FindPatch(cs[i][0], cs[i][1]);

	 /// Add the entity to this patch.
	 if (p)
	 {
	    p->AddEntity (ent);
	    data->m_Patches.push_back (p);
	 }
      }
   }
   
   /*
    * Make a list of potential colliders.
    * Note: we use a map to ensure there no collision is computed
    * twice.
    */
   void
   Quadtree::GetColliders (ColliderList &list)
   {
       m_Rootnode->CollectColliders(list);
   }    


   // =======================================================================

   struct RenderStats
   {
	 int m_NumTris;
	 int m_NumFlushes;
   } g_RenderStats;

   bool g_UpdateVis = true;
   
   // =======================================================================
   // Triangle bin...
   // =======================================================================
   
   TriBin::TriBin (const MaterialList& materials) :
      m_Black ("black"),
      m_CurPatch (0)
   {
      SetMaterials (materials);

      m_VBMaterials.resize(100);

      m_Black.m_Flags = MATERIAL_HAS_PASS1;
      m_Black.m_Passes[0].m_Flags = 0;
   }

   void
   TriBin::SetMaterials (const MaterialList& materials)
   {
       const size_t size = materials.size();

       m_Materials.resize( size );
       m_Triangles.resize( size + 1 );
       size_t i;

       for (i=0 ; i<size ; ++i)
       {
	   m_Materials[i] = materials[i];
	   m_Materials[i]->Ref();
       }

       for (i=0 ; i<=size ; ++i)
       {
	   m_Triangles[i].SetType (PRIM_TRIANGLES);
	   m_Triangles[i].Resize(100);
	   m_Triangles[i].SetEnabledSize(0);
       }

   }

   TriBin::~TriBin ()
   {
      for (size_t i = 0; i < m_Materials.size(); i++)
	 m_Materials[i]->Unref();
   }

   void
   TriBin::AddTriangle (int x1, int y1,
			int x2, int y2,
			int x3, int y3)
   {
      if (m_CurPatch == NULL)
	 return;

      const uchar material1 = m_CurPatch->Ground (x1, y1);
      const uchar material2 = m_CurPatch->Ground (x2, y2);
      const uchar material3 = m_CurPatch->Ground (x3, y3);

      const int index1 = m_CurPatch->Index (x1, y1);
      const int index2 = m_CurPatch->Index (x2, y2);
      const int index3 = m_CurPatch->Index (x3, y3);

      // By default, triangles are blended...
      size_t materialnum = m_Materials.size();
      bool mixed = true;

      // ...unless the three vertex have the same material. 
      if ((material1 == material2) &&
	  (material2 == material3))
      {
	 materialnum = material1;
	 mixed = false;
      }

      PrimitiveBlock &pb = m_Triangles[materialnum];
      size_t triangleindex = pb.EnabledSize();

      // Reserve space if necessary...
      if (triangleindex + 3 >= pb.Size())
	 pb.Resize (triangleindex * 2);

      // Put the materials of mixed-material triangles.
      if (mixed)
      {
	 if (triangleindex + 3 >= m_VBMaterials.size())
	    m_VBMaterials.resize (triangleindex*2);

	 m_VBMaterials[triangleindex] = material1;
	 m_VBMaterials[triangleindex+1] = material2;
	 m_VBMaterials[triangleindex+2] = material3;
      }

      // Put indices...
      pb[triangleindex++] = index1;
      pb[triangleindex++] = index2;
      pb[triangleindex++] = index3;

      pb.SetEnabledSize (triangleindex);
   }

   void
   TriBin::FlushTris (Renderer &render,
		      const Material &material,
		      const PrimitiveBlock &pb,
		      TriType type)
   {
      Material *mat = (Material*) &material;
      int flags = mat->m_Flags;
      int pflags = mat->m_Passes[0].m_Flags;

      g_RenderStats.m_NumTris += pb.EnabledSize() / 3;
      
      if (type != BLEND)
	 mat->m_Passes[0].m_Flags &= ~PASS_HAS_BLENDING;
      else
      {
	 mat->m_Flags &= ~MATERIAL_HAS_PASS2;
	 mat->m_Flags &= ~MATERIAL_HAS_PASS3;
	 mat->m_Flags &= ~MATERIAL_HAS_PASS4;
      }
            
      render.RenderBlock (*mat, pb.Type(), &pb[0], pb.EnabledSize());
      
      mat->m_Flags = flags;
      mat->m_Passes[0].m_Flags = pflags;
   }

   void
   TriBin::Flush (Renderer &render)
   {
      ++g_RenderStats.m_NumFlushes;
      
      /// Initialize the vertex buffer.
      render.SetActiveVB (m_CurPatch->m_VB);
      render.LockVB (0, m_CurPatch->m_VB.Size());

      // Mixed terrain triangles (ie which will need blending)
      PrimitiveBlock &pb = m_Triangles.back();

      // Render 'em to black so additive blending will work.
      if (pb.EnabledSize() != 0)
      {
 	 glDisable (GL_COLOR_ARRAY);
	 glDisable (GL_CULL_FACE);

	 glColor4f (0.0f, 0.0f, 0.0f, 1.0f);
	 FlushTris(render, m_Black, pb, NORMAL);
	 glColor4f (1.0f, 1.0f, 1.0f, 1.0f);

	 glEnable (GL_CULL_FACE);
	 glEnable (GL_COLOR_ARRAY);
      }
      
      const size_t materialCount = m_Materials.size();
      
      for (size_t j = 0; j < materialCount; j++)
      {
	 // Draw the "normal" non-blended triangles ('one material' tris)
	 FlushTris (render, *m_Materials[j], m_Triangles[j], NORMAL);
      }
      
      glColorMaterial(GL_FRONT, GL_DIFFUSE);
      glEnable (GL_COLOR_MATERIAL);
      if (pb.EnabledSize())
      {
	 for (size_t j = 0; j < materialCount; j++)
	 {
	    bool flush = false;

	    // Set alpha values : the alpha value at a corner will be max
	    // if this corner is of the same material than the one we're 
	    // drawing. The alpha values of other corners will be zero, so
	    // triangles will blend smoothly from one material to another
	    // (ie. bilinear filtering)
	    for (size_t i = 0; i < pb.EnabledSize(); i++)
	    {
	       RGBA &color = m_CurPatch->m_VB.Color4 (pb[i]);

	       if (m_VBMaterials[i] == j) 
	       {
		  color.A = 255;
		  flush = true;
	       }
	       else
		  color.A = 0;
	    }
	    

	    if (!flush)
	       continue;

	    // Render 'multiple terrain' triangles with blending.
	    render.OverrideVB ( VertexBuffer::VB_HAS_COLOR, m_CurPatch->m_VB);
	    FlushTris (render, *m_Materials[j],  pb, BLEND);
	 }
      }
      glColor4f (1.0, 1.0, 1.0, 1.0);
      glDisable (GL_COLOR_MATERIAL);
      
      render.UnlockVB();
	
      for (size_t i = 0; i < m_Triangles.size(); i++)
	 m_Triangles[i].SetEnabledSize(0);
   }

   /////////////////////////////////////////////////////////////////////

   void
   QuadtreeNode::Render (
		     Renderer &render,
		     TriBin &bin,
		     Visibility vis)
   {
      if (g_UpdateVis)
	  m_Vis = vis;

      if (m_Vis != INSIDE)
      {
	  if (g_UpdateVis)
	      m_Vis = render.GetFrustum().GetVisibility (m_BBox);

	  if (m_Vis == OUTSIDE )
	      return;
      }
      
      // We assume a patch do not have children
      if (!m_Children[0])
      {
	  Patch *patch  = static_cast< Patch* >( this );

	  if (patch->m_Dirty)
	      patch->Validate();

	  bin.SetPatch (patch);

	  for (int z=0 ; z<Patch::SIZE ; ++z)
	  {
	      const int mz0 = patch->m_MinZ + z;
	      const int mz1 = mz0 + 1;
	      for (int x=0 ; x<Patch::SIZE ; ++x)
	      {
		  const int mx0 = patch->m_MinX + x;
		  const int mx1 = mx0 + 1;
		  bin.AddTriangle (mx0, mz0, mx1, mz0, mx1, mz1);
		  bin.AddTriangle (mx1, mz1, mx0, mz1, mx0, mz0);
	      }
	  }

	  bin.Flush(render);
	  bin.SetPatch (NULL);
      }
      else
      {
	  for (int i = 0; i < 4; i++)
	  {
	      m_Children[i]->Render (render, bin, (Visibility)m_Vis);
	  }
      }
   }

   /////////////////////////////////////////////////////////////////////

   void
   Quadtree::Render (Renderer &renderer, const Camera &camera)
   {
      g_RenderStats.m_NumTris = 0;
      g_RenderStats.m_NumFlushes = 0;

      m_Rootnode->Render (renderer, m_TriBin, SOME_CLIP);

#if 0
      Sys()->Log ("Quadtree rendered : %d triangles, %d flushes\n", 
		  g_RenderStats.m_NumTris,
		  g_RenderStats.m_NumFlushes);
#endif
   }

		      
   
} // ns Ark
