/* $Id: HFQuadtree.h,v 1.6 2002/12/27 11:56:00 mrq Exp $
**
** Ark - Libraries, Tools & Programs for MMORPG developpements.
** Copyright (C) 1999-2000 The Contributors of the Ark Project
** Copyright (C) 1999 Amit J. Patel
** Please see the file "AUTHORS" for a list of contributors
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#ifndef ENGINE_QUADTREE_H
#define ENGINE_QUADTREE_H

#include <vector>
#include <map>

#include <Ark/Ark.h>
#include <Ark/ArkMath.h>
#include <Ark/ArkRay.h>
#include <Ark/ArkVBuffer.h>

#include <Modules/HeightField/ArkHeightField.h>
#include <Modules/HeightField/HFWorld.h>

namespace Ark
{
   class Patch;
   class QuadtreeNode;
 
   /**
    * Those are height-field specific entity data, a list of patches this
    * entity belongs to. This structure is created when the entity is added
    * to the world, it is updated when the entity moves, and destroyed when
    * the entity is removed of the world.
    * It is stored in the m_WorldData member of the entity.
    */
   class ARKHEIGHTFIELD_DLL_API HFEntityData
   {
       public:
	   BBox m_OldBbox;
	   std::vector< Patch* > m_Patches;
   };

   // Note: we use a map to ensure there no collision is computed
   //       twice.
   struct ARKHEIGHTFIELD_DLL_API Collider
   {
	 Entity *m_E1, *m_E2;

	 Collider (Entity *e1, Entity *e2)
	 {
	    assert (e1 != e2);

	    if (e1 > e2)
	    {
	       m_E1 = e1;
	       m_E2 = e2;
	    }
	    else
	    {
	       m_E1 = e2;
	       m_E2 = e1;
	    }
	 }
   };

   ARKHEIGHTFIELD_DLL_API bool operator < (const Collider &b, const Collider &c);
   ARKHEIGHTFIELD_DLL_API bool operator == (const Collider &b, const Collider &c);

   typedef std::map<Collider, bool> ColliderList;

   /** 
    *
    */
   class TriBin
   {
	 Material m_Black;
	 MaterialList m_Materials;
	 BlockList m_Triangles;
	 std::vector<size_t> m_VBMaterials;

	 Patch *m_CurPatch;
	
	 enum TriType {NORMAL,BLEND};
	 void FlushTris (Renderer &render,
			 const Material &material,
			 const PrimitiveBlock &pb,
			 TriType type);
			 

      public:
	 TriBin (const MaterialList& materials);
	 ~TriBin ();

	 void SetMaterials (const MaterialList& materials);

	 void SetPatch (Patch *patch = NULL)
	 {m_CurPatch = patch;}

	 void AddTriangle (int x1, int y1,
			   int x2, int y2,
			   int x3, int y3);
	 void Flush (Renderer &render);
   };


   /** 
    *
    */
   class ARKHEIGHTFIELD_DLL_API QuadtreeNode
   {
      public:
	 QuadtreeNode();
	 virtual ~QuadtreeNode ();

	 virtual void Build(HeightField* hf, int mx, int mz, int size);
	 void Render (Renderer &render, TriBin &bin, Visibility vis);
	 Patch* FindPatch(scalar x, scalar z);
	 void InvalidatePatches(const BBox& box);
	 virtual void CollectColliders (ColliderList &list);
	 inline const BBox& GetBoundingBox() const { return m_BBox; };


	 bool RayTrace (const Ray &ray,
			std::vector<int>& potentialtris,
			EntityList& potentialents);

      protected:
	 QuadtreeNode *m_Children[4];
	 BBox m_BBox;

	 uchar m_Vis;
   };

   /** 
    *
    */
   class ARKHEIGHTFIELD_DLL_API Patch : public QuadtreeNode
   {	    
      private:
	 friend class TriBin;
	 friend class QuadtreeNode;
	 
	 VertexBuffer m_VB;
	 HeightField *m_Heightfield;
	 
	 EntityList m_Entities;
	 
      public:
	 //enum {SIZE=32};
	 enum { SIZE=16 };

	 // I do not want them to be recomputed every frame
	 int m_MinX, m_MinZ;

	 Material m_Black;
	 MaterialList m_Materials;
	 BlockList m_Triangles;
	 std::vector<size_t> m_VBMaterials;

	 Vector3 m_Offset;

	 bool m_Dirty;
	 bool m_CollisionDirty;

      public:
	 Patch();
	 virtual ~Patch ();
	 
	 virtual void Build(HeightField* hf, int mx, int mz, int size);
	 void Validate();

	 uchar Ground (int x, int y) const;
	 int Index (int x, int y) const;

	 /// Add an entity to this patch..
	 void AddEntity (Entity *ent);
	 
	 /// Remove an entity from this patch..
	 void RemoveEntity (Entity *ent);

	 /**
	  * Add all the patches entities to the provided list, checking that no
	  * entity is added twice in the list.
	  */
	 void AddPatchEntsToList(EntityList &list);

	 void RayTrace (const Ray &ray, std::vector<int>& potentialtris);

	 /**
	  * Make a list of potential colliders. If nothing has moved
	  * in the patch, we assume there can't be any collision.
	  * Note: we use a map to ensure there no collision is computed
	  * twice.
	  * Once this function has returned, the patch collision dirty
	  * flag is cleared.
	  */
	 virtual void CollectColliders (ColliderList &list);
	 
      private:
	 Vector3 ComputeNormal (int x, int z, const Vector3& coord) const;

   };

   /**
    * A quadtree...
    */
   class ARKHEIGHTFIELD_DLL_API Quadtree
   {
	 HeightField *m_Heightfield;
	 TriBin m_TriBin;
	 QuadtreeNode *m_Rootnode;
	 int m_Size;

      public:
	 Quadtree (HeightField *heightfield);
	 ~Quadtree ();
	 
	 void Render (Renderer &renderer, const Camera& camera);

	 /**
	  * Mark the patches in the specified area as "invalidate", which
	  * means their topology has changed since the quadtree has been
	  * created.
	  */
	 void Invalidate(scalar minx, scalar minz, scalar maxx, scalar maxz);

	 void SetMaterials (const MaterialList& materials) 
	 { m_TriBin.SetMaterials( materials ); }
	 
	 // Trace a ray through the world trying to detect a collision...
	 bool RayTrace (const Ray &ray, Collision &collision, bool test_ents);

	 // Remove an entity from the quadtree's patches...
	 void RemoveEntity (Entity *ent, HFEntityData *data);

	 // Add/update an entity in the quadtree's patches.
	 void UpdateEntity (Entity *ent, HFEntityData *data);

	 /**
	  * Make a list of potential colliders.
	  * Note: we use a map to ensure there no collision is computed
	  * twice.
	  */
	 void GetColliders (ColliderList &list);
   };


} // namespace Ark


#endif // QUADTREE_HPP
